@@ -593,12 +593,9 @@ ST_FUNC void store(int r, SValue *sv)
 static void arm64_gen_bl_or_b(int b)
 {
     if ((vtop->r & (VT_VALMASK | VT_LVAL)) == VT_CONST) {
-        assert(!b);
-        if (vtop->r & VT_SYM)
-            greloc(cur_text_section, vtop->sym, ind, R_AARCH64_CALL26);
-        else
-            assert(0);
-        o(0x94000000); // bl .
+        assert(!b && (vtop->r & VT_SYM));
+	greloc(cur_text_section, vtop->sym, ind, R_AARCH64_CALL26);
+	o(0x94000000); // bl .
     }
     else
         o(0xd61f0000 | (uint32_t)!b << 21 | intr(gv(RC_R30)) << 5); // br/blr
@@ -1826,7 +1823,10 @@ ST_FUNC void gen_vla_sp_save(int addr) {
 }
 
 ST_FUNC void gen_vla_sp_restore(int addr) {
-    uint32_t r = intr(get_reg(RC_INT));
+    // Use x30 because this function can be called when there
+    // is a live return value in x0 but there is nothing on
+    // the value stack to prevent get_reg from returning x0.
+    uint32_t r = 30;
     arm64_ldrx(0, 3, r, 29, addr);
     o(0x9100001f | r << 5); // mov sp,x(r)
 }
