@@ -21,7 +21,7 @@
 /* tccgen: asm_label cleanu */
 //#define BEFORE_hash_opt
 
-#define FUNC_STR func_str.str
+#define FUNC_STR func_str->str
 
 /*****************************************************************************/
 /*                            exsymtab_token_hash                            */
@@ -691,9 +691,9 @@ Sym * get_new_symtab_pointer (Sym * old, ram_hash * rh)
         to_return->type.t = old->type.t;
 
     /* The type.ref field contains something useful only if the basic type
-     * is a pointer, struct, or function. See code from tccgen's
+     * is a pointer, struct, enum, or function. See code from tccgen's
      * compare_types for details. */
-    if (btype == VT_PTR || btype == VT_STRUCT || btype == VT_FUNC) {
+    if (btype == VT_PTR || btype == VT_STRUCT || btype == VT_ENUM || btype == VT_FUNC) {
         to_return->type.ref = get_new_symtab_pointer(old->type.ref, rh);
     }
 
@@ -891,6 +891,7 @@ void copy_extended_symtab (TCCState * s, Sym * define_start, int tok_start)
              * for contrast) */
 
             ts_len = tokenstream_len(old_func->FUNC_STR);
+            new_func->func_str = tcc_malloc(sizeof(TokenString));
             new_func->FUNC_STR = tcc_malloc(ts_len * sizeof(int));
             memcpy(new_func->FUNC_STR, old_func->FUNC_STR, ts_len * sizeof(int));
             to_return->inline_funcs[i] = new_func;
@@ -997,6 +998,7 @@ LIBTCCAPI void tcc_delete_extended_symbol_table (extended_symtab * symtab)
         int i;
         for (i = 0; i < symtab->N_inline_funcs; i++) {
             tcc_free(symtab->inline_funcs[i]->FUNC_STR);
+            tcc_free(symtab->inline_funcs[i]->func_str);
             tcc_free(symtab->inline_funcs[i]);
         }
         tcc_free(symtab->inline_funcs);
@@ -1288,6 +1290,7 @@ void copy_extended_tokensym (extended_symtab * symtab, TokenSym * from, TokenSym
              * for contrast) */
 
             ts_len = tokenstream_len(old_func->FUNC_STR);
+            new_func->func_str = tcc_malloc(sizeof(TokenString));
             new_func->FUNC_STR = tcc_malloc(ts_len * sizeof(int));
             memcpy(new_func->FUNC_STR, old_func->FUNC_STR, ts_len * sizeof(int));
             tokenstream_copy(old_func->FUNC_STR, new_func->FUNC_STR, symtab);
@@ -1356,13 +1359,14 @@ void copy_extended_tokensym (extended_symtab * symtab, TokenSym * from, TokenSym
 }
 
 /* Copy the CType information from an extended sym into a local CType. The hard
- * part here is finding the local tokensym associated with the type.ref field,
- * which is only an issue if the type is a pointer, struct, or function. */
+ * part here is finding the local tokensym associated with the type.ref field.
+ * In principle this is only an issue if the type is a pointer, struct, enum,
+ * or function, but at this point checking non-null is sufficient. */
 
 #define copy_ctype(to_type, from, symtab) do { \
     int btype = from->type.t & VT_BTYPE; \
     to_type.t = from->type.t; \
-    if (btype == VT_PTR || btype == VT_STRUCT || btype == VT_FUNC) { \
+    if (from->type.ref != NULL) { \
         /* Get the from->type.ref's token and look for it here */ \
         if (from->type.ref->v & SYM_FIRST_ANOM) { \
             /* Anonymous symbol; just copy it. */ \
@@ -1693,10 +1697,10 @@ int exsymtab_deserialize_token_stream(FILE * in_fh, Sym * curr_sym, int i)
 
 int exsymtab_serialize_type_ref(FILE * out_fh, Sym * curr_sym, ram_hash * offset_rt)
 {
-    /* For details, see notes under copy_extended_symtab */
+    /* Get offset if non-null. */
     int btype = curr_sym->type.t & VT_BTYPE;
-    void * to_write;
-    if (btype == VT_PTR || btype == VT_STRUCT || btype == VT_FUNC) {
+    void * to_write = NULL;
+	if (curr_sym->type.ref != NULL) {
         /* write the offset */
         void ** p_offset = ram_hash_get_ref(offset_rt, curr_sym->type.ref);
         to_write = *p_offset;
@@ -2218,14 +2222,23 @@ int exsymtab_deserialize_inline_func_token_stream(FILE * in_fh,
         return 0;
     }
 
+    curr_func->func_str = tcc_malloc(sizeof(InlineFunc));
+    if (curr_func->func_str == NULL) {
+        printf("Deserialization failed: Unable to allocate memory for "
+            "Inline function number %d\n", inline_offset);
+        return 0;
+	}
+    
     stream = tcc_malloc(sizeof(int) * len);
     if (stream == NULL) {
+		tcc_free(curr_func->func_str);
         printf("Deserialization failed: Unable to allocate memory for "
             "token stream for Inline function number %d\n", inline_offset);
         return 0;
     }
 
     if (fread(stream, sizeof(int), len, in_fh) != len) {
+		tcc_free(curr_func->func_str);
         tcc_free(stream);
         printf("Deserialization failed: Unable to get token stream for "
             "Inline function number %d\n", inline_offset);
