@@ -23,40 +23,51 @@
 /* only native compiler supports -run */
 #ifdef TCC_IS_NATIVE
 
+#ifndef _WIN32
+# include <sys/mman.h>
+#endif
+
 #ifdef CONFIG_TCC_BACKTRACE
+# ifndef _WIN32
+#  include <signal.h>
+#  ifndef __OpenBSD__
+#   include <sys/ucontext.h>
+#  endif
+# else
+#  define ucontext_t CONTEXT
+# endif
 ST_DATA int rt_num_callers = 6;
 ST_DATA const char **rt_bound_error_msg;
 ST_DATA void *rt_prog_main;
-#endif
-
-#ifdef _WIN32
-#define ucontext_t CONTEXT
+static int rt_get_caller_pc(addr_t *paddr, ucontext_t *uc, int level);
+static void rt_error(ucontext_t *uc, const char *fmt, ...);
+static void set_exception_handler(void);
 #endif
 
 static void set_pages_executable(void *ptr, unsigned long length);
-static void set_exception_handler(void);
-static int rt_get_caller_pc(addr_t *paddr, ucontext_t *uc, int level);
-static void rt_error(ucontext_t *uc, const char *fmt, ...);
 static int tcc_relocate_ex(TCCState *s1, void *ptr);
 
 #ifdef _WIN64
-static void win64_add_function_table(TCCState *s1);
+static void *win64_add_function_table(TCCState *s1);
+static void win64_del_function_table(void *);
 #endif
 
+// #define HAVE_SELINUX
+
 /* ------------------------------------------------------------- */
 /* Do all relocations (needed before using tcc_get_symbol())
    Returns -1 on error. */
 
 LIBTCCAPI int tcc_relocate(TCCState *s1, void *ptr)
 {
-    int ret;
+    int size;  void *mem;
 
     if (TCC_RELOCATE_AUTO != ptr)
         return tcc_relocate_ex(s1, ptr);
 
-    ret = tcc_relocate_ex(s1, NULL);
-    if (ret < 0)
-        return ret;
+    size = tcc_relocate_ex(s1, NULL);
+    if (size < 0)
+        return -1;
 
 #ifdef HAVE_SELINUX
     {   /* Use mmap instead of malloc for Selinux.  Ref:
@@ -64,39 +75,59 @@ LIBTCCAPI int tcc_relocate(TCCState *s1, void *ptr)
 
         char tmpfname[] = "/tmp/.tccrunXXXXXX";
         int fd = mkstemp (tmpfname);
+        void *wr_mem;
 
-        s1->mem_size = ret;
         unlink (tmpfname);
-        ftruncate (fd, s1->mem_size);
+        ftruncate (fd, size);
 
-        s1->write_mem = mmap (NULL, ret, PROT_READ|PROT_WRITE,
+        wr_mem = mmap (NULL, size, PROT_READ|PROT_WRITE,
             MAP_SHARED, fd, 0);
-        if (s1->write_mem == MAP_FAILED)
+        if (wr_mem == MAP_FAILED)
             tcc_error("/tmp not writeable");
-
-        s1->runtime_mem = mmap (NULL, ret, PROT_READ|PROT_EXEC,
+        mem = mmap (NULL, size, PROT_READ|PROT_EXEC,
             MAP_SHARED, fd, 0);
-        if (s1->runtime_mem == MAP_FAILED)
+        if (mem == MAP_FAILED)
             tcc_error("/tmp not executable");
 
-        ret = tcc_relocate_ex(s1, s1->write_mem);
+        tcc_relocate_ex(s1, wr_mem);
+        dynarray_add(&s1->runtime_mem, &s1->nb_runtime_mem, (void*)(addr_t)size);
+        dynarray_add(&s1->runtime_mem, &s1->nb_runtime_mem, wr_mem);
+        dynarray_add(&s1->runtime_mem, &s1->nb_runtime_mem, mem);
     }
 #else
-    s1->runtime_mem = tcc_malloc(ret);
-    ret = tcc_relocate_ex(s1, s1->runtime_mem);
+    mem = tcc_malloc(size);
+    tcc_relocate_ex(s1, mem); /* no more errors expected */
+    dynarray_add(&s1->runtime_mem, &s1->nb_runtime_mem, mem);
 #endif
-    return ret;
+    return 0;
+}
+
+ST_FUNC void tcc_run_free(TCCState *s1)
+{
+    int i;
+
+    for (i = 0; i < s1->nb_runtime_mem; ++i) {
+#ifdef HAVE_SELINUX
+        int size = (int)(addr_t)s1->runtime_mem[i];
+        munmap(s1->runtime_mem[++i], size);
+        munmap(s1->runtime_mem[++i], size);
+#else
+# ifdef _WIN64
+        win64_del_function_table(*(void**)s1->runtime_mem[i]);
+# endif
+        tcc_free(s1->runtime_mem[i]);
+#endif
+    }
+    tcc_free(s1->runtime_mem);
 }
 
 /* launch the compiled program with the given arguments */
 LIBTCCAPI int tcc_run(TCCState *s1, int argc, char **argv)
 {
     int (*prog_main)(int, char **);
-    int ret;
 
     if (tcc_relocate(s1, TCC_RELOCATE_AUTO) < 0)
         return -1;
-
     prog_main = tcc_get_symbol_err(s1, s1->runtime_main);
 
 #ifdef CONFIG_TCC_BACKTRACE
@@ -106,13 +137,15 @@ LIBTCCAPI int tcc_run(TCCState *s1, int argc, char **argv)
     }
 #endif
 
+    errno = 0; /* clean errno value */
+
 #ifdef CONFIG_TCC_BCHECK
     if (s1->do_bounds_check) {
         void (*bound_init)(void);
         void (*bound_exit)(void);
         void (*bound_new_region)(void *p, addr_t size);
         int  (*bound_delete_region)(void *p);
-        int i;
+        int i, ret;
 
         /* set error function */
         rt_bound_error_msg = tcc_get_symbol_err(s1, "__bound_error_msg");
@@ -121,28 +154,24 @@ LIBTCCAPI int tcc_run(TCCState *s1, int argc, char **argv)
         bound_exit = tcc_get_symbol_err(s1, "__bound_exit");
         bound_new_region = tcc_get_symbol_err(s1, "__bound_new_region");
         bound_delete_region = tcc_get_symbol_err(s1, "__bound_delete_region");
+
         bound_init();
         /* mark argv area as valid */
         bound_new_region(argv, argc*sizeof(argv[0]));
         for (i=0; i<argc; ++i)
-            bound_new_region(argv[i], strlen(argv[i]));
+            bound_new_region(argv[i], strlen(argv[i]) + 1);
 
-	errno = 0; /* clean errno value */
         ret = (*prog_main)(argc, argv);
 
         /* unmark argv area */
         for (i=0; i<argc; ++i)
             bound_delete_region(argv[i]);
         bound_delete_region(argv);
-
         bound_exit();
-    } else
-#endif
-    {
-	errno = 0; /* clean errno value */
-        ret = (*prog_main)(argc, argv);
+        return ret;
     }
-    return ret;
+#endif
+    return (*prog_main)(argc, argv);
 }
 
 /* relocate code. Return -1 on error, required size if ptr is NULL,
@@ -169,15 +198,17 @@ static int tcc_relocate_ex(TCCState *s1, void *ptr)
     }
 
     offset = 0, mem = (addr_t)ptr;
+#ifdef _WIN64
+    offset += sizeof (void*);
+#endif
     for(i = 1; i < s1->nb_sections; i++) {
         s = s1->sections[i];
         if (0 == (s->sh_flags & SHF_ALLOC))
             continue;
-        length = s->data_offset;
-        s->sh_addr = mem ? (mem + offset + 15) & ~15 : 0;
-        offset = (offset + length + 15) & ~15;
+        offset = (offset + 15) & ~15;
+        s->sh_addr = mem ? mem + offset : 0;
+        offset += s->data_offset;
     }
-    offset += 16;
 
     /* relocate symbols */
     relocate_syms(s1, 1);
@@ -195,6 +226,10 @@ static int tcc_relocate_ex(TCCState *s1, void *ptr)
     }
     relocate_plt(s1);
 
+#ifdef _WIN64
+    *(void**)ptr = win64_add_function_table(s1);
+#endif
+
     for(i = 1; i < s1->nb_sections; i++) {
         s = s1->sections[i];
         if (0 == (s->sh_flags & SHF_ALLOC))
@@ -211,11 +246,6 @@ static int tcc_relocate_ex(TCCState *s1, void *ptr)
             set_pages_executable(ptr, length);
     }
 
-#ifdef _WIN64
-// XXX how badly does this removal break Windows things?
-//    win64_add_function_table(s1);
-#endif
-
 /* #ifdef CONFIG_TCC_EXSYMTAB */
     /* If they have an extended symbol table, copy the symbol pointers. */
     if (s1->exsymtab > (extended_symtab*)1) copy_extended_symbols_to_exsymtab(s1);
@@ -233,7 +263,6 @@ static void set_pages_executable(void *ptr, unsigned long length)
     unsigned long old_protect;
     VirtualProtect(ptr, length, PAGE_EXECUTE_READWRITE, &old_protect);
 #else
-    extern void __clear_cache(void *beginning, void *end);
 #ifndef PAGESIZE
 # define PAGESIZE 4096
 #endif
@@ -241,15 +270,39 @@ static void set_pages_executable(void *ptr, unsigned long length)
     start = (addr_t)ptr & ~(PAGESIZE - 1);
     end = (addr_t)ptr + length;
     end = (end + PAGESIZE - 1) & ~(PAGESIZE - 1);
-    mprotect((void *)start, end - start, PROT_READ | PROT_WRITE | PROT_EXEC);
-  #ifndef __PCC__
-    __clear_cache(ptr, (char *)ptr + length);
-  #else
-    /* pcc 1.2.0.DEVEL 20141206 don't have such proc */
-  #endif
+    if (mprotect((void *)start, end - start, PROT_READ | PROT_WRITE | PROT_EXEC))
+        tcc_error("mprotect failed: did you mean to configure --with-selinux?");
+#if defined TCC_TARGET_ARM || defined TCC_TARGET_ARM64
+    { extern void __clear_cache(void *beginning, void *end);
+      __clear_cache(ptr, (char *)ptr + length); }
+#endif
 #endif
 }
 
+#ifdef _WIN64
+static void *win64_add_function_table(TCCState *s1)
+{
+    void *p = NULL;
+    if (s1->uw_pdata) {
+        p = (void*)s1->uw_pdata->sh_addr;
+        RtlAddFunctionTable(
+            (RUNTIME_FUNCTION*)p,
+            s1->uw_pdata->data_offset / sizeof (RUNTIME_FUNCTION),
+            text_section->sh_addr
+            );
+        s1->uw_pdata = NULL;
+    }
+    return p;;
+}
+
+static void win64_del_function_table(void *p)
+{
+    if (p) {
+        RtlDeleteFunctionTable((RUNTIME_FUNCTION*)p);
+    }
+}
+#endif
+
 /* ------------------------------------------------------------- */
 #ifdef CONFIG_TCC_BACKTRACE
 
@@ -507,6 +560,8 @@ static int rt_get_caller_pc(addr_t *paddr, ucontext_t *uc, int level)
         *paddr = uc->uc_mcontext.eip;
 #elif defined(__NetBSD__)
         *paddr = uc->uc_mcontext.__gregs[_REG_EIP];
+#elif defined(__OpenBSD__)
+        *paddr = uc->sc_eip;
 #else
         *paddr = uc->uc_mcontext.gregs[REG_EIP];
 #endif
@@ -520,6 +575,8 @@ static int rt_get_caller_pc(addr_t *paddr, ucontext_t *uc, int level)
         fp = uc->uc_mcontext.ebp;
 #elif defined(__NetBSD__)
         fp = uc->uc_mcontext.__gregs[_REG_EBP];
+#elif defined(__OpenBSD__)
+        *paddr = uc->sc_ebp;
 #else
         fp = uc->uc_mcontext.gregs[REG_EBP];
 #endif
@@ -684,17 +741,6 @@ static void set_exception_handler(void)
     SetUnhandledExceptionFilter(cpu_exception_handler);
 }
 
-#ifdef _WIN64
-static void win64_add_function_table(TCCState *s1)
-{
-    RtlAddFunctionTable(
-        (RUNTIME_FUNCTION*)s1->uw_pdata->sh_addr,
-        s1->uw_pdata->data_offset / sizeof (RUNTIME_FUNCTION),
-        text_section->sh_addr
-        );
-}
-#endif
-
 /* return the PC at frame level 'level'. Return non zero if not found */
 static int rt_get_caller_pc(addr_t *paddr, CONTEXT *uc, int level)
 {
@@ -759,7 +805,7 @@ static TCCSyms tcc_syms[] = {
     { NULL, NULL },
 };
 
-ST_FUNC void *resolve_sym(TCCState *s1, const char *symbol)
+ST_FUNC void *dlsym(void *handle, const char *symbol)
 {
     TCCSyms *p;
     p = tcc_syms;
@@ -771,13 +817,6 @@ ST_FUNC void *resolve_sym(TCCState *s1, const char *symbol)
     return NULL;
 }
 
-#elif !defined(_WIN32)
-
-ST_FUNC void *resolve_sym(TCCState *s1, const char *sym)
-{
-    return dlsym(RTLD_DEFAULT, sym);
-}
-
 #endif /* CONFIG_TCC_STATIC */
 #endif /* TCC_IS_NATIVE */
 /* ------------------------------------------------------------- */
